(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2013-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: gen_identity.ml,v 1.5 2014-02-26 11:38:45 fclement Exp $ *)

open Gen_identity_model;;

(* Wrappers. *)

let mk_init_proj { Udebug.spf_report = report; _; } vals =
  let nargs = Uarray.length vals in
  if nargs <> Nat._1_ then
    i_failwith "init_proj" "initialization should receive one value" else
  report "one value received";
  let dim = Nat.of_int (Typed_value.to_int vals.(0)) in
  report "expected dimension extracted";
  dim
;;

let mk_init_inj _spf ((ni, nj), ppm1, maa1, j1) =
  let sz = (Nat.to_int ni, Nat.to_int nj) in
  let ppm1 = Umatrix.map Nat.to_float ppm1 in
  let (lens1, a1) = Uarray_ng.array_array_pack maa1 in
  let lens1 = Array.map Nat.to_int lens1 in
  [| Typed_value.of_int_couple sz;
     Typed_value.of_c_float_matrix ppm1;
     Typed_value.of_int_vector lens1;
     Typed_value.of_float_vector a1;
     Typed_value.of_float j1; |]
;;

let make_init_wrapper =
  Pio.make_task_wrapper "init" mk_init_inj mk_init_proj
;;

let mk_grad_proj { Udebug.spf_report = report; _; } vals =
  let nargs = Uarray.length vals in
  if nargs <> Nat._1_ then
    i_failwith "grad_proj" "gradient should receive one value" else
  report "one value received";
  let pm = Typed_value.to_c_float_matrix vals.(0) in
  report "fine parameter matrix extracted";
  pm
;;

let mk_grad_inj _spf gm = [| Typed_value.of_c_float_matrix gm; |];;

let make_grad_wrapper =
  Pio.make_task_wrapper "grad" mk_grad_inj mk_grad_proj
;;

let mk_optim_proj { Udebug.spf_report = report; _; } vals =
  let nargs = Uarray.length vals in
  if nargs <> Nat._3_ then
    i_failwith "optim_proj" "optimization should receive three values" else
  report "three values received";
  let ppm = Typed_value.to_c_float_matrix vals.(0) in
  let ppm = Umatrix.map Nat.of_float ppm in
  report "partition extracted";
  let lens_init = Typed_value.to_int_vector vals.(1) in
  let lens_init = Array.map Nat.of_int lens_init in
  report "initial coarse parameter lengths extracted";
  let a_init = Typed_value.to_float_vector vals.(2) in
  let maa_init = Uarray_ng.array_array_unpack lens_init a_init in
  report "initial coarse parameter extracted";
  (ppm, maa_init)
;;

let mk_optim_inj _spf (maa_opt, j_opt) =
  let (lens_opt, a_opt) = Uarray_ng.array_array_pack maa_opt in
  let lens_opt = Array.map Nat.to_int lens_opt in
  [| Typed_value.of_int_vector lens_opt;
     Typed_value.of_float_vector a_opt;
     Typed_value.of_float j_opt; |]
;;

let make_optim_wrapper =
  Pio.make_task_wrapper "optim" mk_optim_inj mk_optim_proj
;;

let mk_final_proj { Udebug.spf_report = report; _; } vals =
  let nargs = Uarray.length vals in
  if nargs <> Nat._2_ then
    i_failwith "final_proj" "finalization should receive two values" else
  report "two values received";
  let dname = Typed_value.to_string vals.(0) in
  report "final image directory name extracted";
  let pm = Typed_value.to_c_float_matrix vals.(1) in
  report "fine parameter matrix extracted";
  (dname, pm)
;;

let mk_final_inj _spf _ = [| Typed_value.of_string "\nOK, bye!\n"; |];;

let make_final_wrapper =
  Pio.make_task_wrapper "final" mk_final_inj mk_final_proj
;;

(* Task computation functions. *)

let make_initialize get_dim dm_of_img pf piob vals =
  Pio.worker_wrap_task piob
    (make_init_wrapper pf "init") (make_init get_dim dm_of_img) vals
;;

let compute_gradient pf piob vals =
  Pio.worker_wrap_task piob (make_grad_wrapper pf "grad") grad vals
;;

let optimize pf piob vals =
  Pio.worker_wrap_task piob (make_optim_wrapper pf "optim") optim vals
;;

let make_finalize img_of_pm pf piob vals =
  Pio.worker_wrap_task piob
    (make_final_wrapper pf "final") (make_final img_of_pm) vals;
  (raise Exit : unit);
;;

(* Main loop. *)

let welcome_action () = "Start identity model"
and farewell_action () = "Finish identity model"
;;

let mk_task_tbl get_dim dm_of_img img_of_pm =
  Ulist1.of_list
    [
     ("init", make_initialize get_dim dm_of_img);
     ("grad", compute_gradient);
     ("optim", optimize);
     ("final", make_finalize img_of_pm);
    ]
;;

let make_main get_dim dm_of_img img_of_pm () =
  let task_tbl = mk_task_tbl get_dim dm_of_img img_of_pm in
  Pio.worker_main welcome_action farewell_action task_tbl
;;
